# <path to game directory>/addons/eventscripts/_engines/python/Lib/esutils/
# commands/__init__.py
# by Adam Cunnington

"""Provide a powerful API to facilitate the registering of client console, 
client say and server commands. Also, give methods to manage access 
to client commands and catch arguments passed with commands.

"""

import shlex

from esutils import admins, tools
import es

__all__ = (
    "ArgumentCountError", 
    "ArgumentValueError", 
    "CommandsError", 
    "Callback", 
    "ClientConsoleCommand", 
    "ClientSayCommand", 
    "Parameter", 
    "ServerCommand", 
)

class ArgumentCountError(Exception):
    """Invalid amount of arguments passed."""
    
class ArgumentValueError(Exception):
    """Invalid value for an argument passed."""
    
class CommandsError(Exception):
    """General error encountered relating to admin.libraries.commands"""
    
class _CommandManager(object):
    def __init__(self):
        self.client_console_commands = {}
        self.client_say_commands = {}
        self.server_commands = {}
        
_command_manager = _CommandManager()

class _Command(object):
    REGISTER_CALLABLE = None
    UNREGISTER_CALLABLE = None
    
    def __init__(self, callback):
        """Instantiate a _Command object.
        
        Arguments:
        callback - the callback object to use.
        
        """
        self._callback = callback
        self.command_names = set()
        
    def _check(self, name):
        if name in self.MAPPING:
            raise CommandsError("%s is already registered as a %s" 
                                %(name, self.VERBOSE_NAME))
        return True
        
    def _invalid_argument(self, parameter, value, user_id=None):
        message = ("ArgumentValueError: %s is an invalid value for parameter, "
                   "%s" %(value, parameter.basename))
        if user_id is None:
            self.INFORMER(message)
        else:
            self.INFORMER(user_id, message) 
            
    def _invalid_syntax(self, user_id=None):
        message = "ArgumentCountError: Parameter Syntax: %s" % self.syntax
        if user_id is None:
            self.INFORMER(message)
        else:
            self.INFORMER(user_id, message) 
            
    def _no_auth(self, user_id=None):
        message = ("You do not have the required permission to use this "
                   "command")
        if user_id is None:
            self.INFORMER(message)
        else:
            self.INFORMER(user_id, message) 
            
    def register(self, name):
        """Register a command name to trigger the command's callback.
        
        Arguments:
        name - the command name.
        
        """
        if "\"" in name:
            raise CommandsError("command words must not contain argument "
                                "delimiters.")
        if self._check(name):
            self.command_names.add(name)
            if (self.REGISTER_CALLABLE is not None and 
                not es.exists(self.BASENAME, name)):
                self.REGISTER_CALLABLE(name, "_commands/%s" 
                                             % self.INTERNAL_CALLBACK.__name__)
            self.MAPPING[name] = self
            
    def unregister(self):
        """Unregister all registered command names."""
        if self.UNREGISTER_CALLABLE is None:
            self.command_names.clear()
            return
        while self.command_names:
            command_name = self.command_names.pop()
            if es.exists(self.BASENAME, command_name):
                self.UNREGISTER_CALLABLE(command_name)
            del self.MAPPING[command_name]
            
    @property
    def syntax(self):
        syntax = ""
        for parameter in self._callback.parameters:
            if parameter in self._callback.defaults:
                syntax += " [%s=%s]" %(parameter.basename, 
                                       self._callback.defaults[parameter])
            else:
                syntax += " <%s>" % parameter.basename
        return syntax
        
class Callback(object):
    """Create a callback that can take a number of parameters and can be used 
    by several command types.
    
    """

    def __init__(self, callable_name):
        """Instantiate a Callback object.
        
        Arguments:
        callable_name - the name to call when a command is submitted with 
        valid syntax and argument values.
        
        """
        self.callable_name = callable_name
        self.parameters = []
        self.defaults = {}
        
    def set_parameter(self, parameter, optional=False, default=None):
        """Set a parameter for the callback.
        
        Arguments:
        parameter - the parameter to register for the callback.
        optional (Keyword Default: False) - whether or not the parameter is 
        optional.
        default (Keyword Default: None) - if the parameter is optional, the 
        value to be sent to the parameter if the argument is omitted.
        
        """
        self.parameters.append(parameter)
        if not optional:
            return
        self.defaults[parameter] = default
        
def _client_console_command():
    user_id = es.getcmduserid()
    command = _command_manager.client_console_commands[es.getargv(0)]
    if (command.requires_auth and 
        not admins.is_user_authorised(es.getplayersteamid(user_id), 
                                      command.permission)):
        command._no_auth(user_id)
        return
    args = [es.getargv(arg) for arg in xrange(1, es.getargc())]
    try:
        args = _get_transformed_args(command._callback, args)
    except ArgumentCountError:
        command._invalid_syntax(user_id)
    except ArgumentValueError, (parameter, value):
        command._invalid_argument(parameter, value, user_id=user_id)
    else:
        command._callback.callable_name(command.INFORMER, user_id=user_id, 
                                       *args)
        
es.addons.registerBlock("_commands", "_client_console_command", 
                        _client_console_command)

class ClientConsoleCommand(_Command):
    """Manage a client console command, wrapping es.regclientcmd"""
    
    BASENAME = "clientcommand"
    INFORMER = staticmethod(tools.echo)
    INTERNAL_CALLBACK = _client_console_command
    MAPPING = _command_manager.client_console_commands
    REGISTER_CALLABLE = staticmethod(es.regclientcmd)
    UNREGISTER_CALLABLE = staticmethod(es.unregclientcmd)
    VERBOSE_NAME = "Client Console Command"
    
    def __init__(self, *args, **kwargs):
        """Instantiate a ClientConsoleCommand object.
        
        Arguments:
        args - the positional arguments defined by the parent class.
        kwargs - the keyword arguments defined by the parent class.
        
        """
        super(ClientConsoleCommand, self).__init__(*args, **kwargs)
        self.requires_auth = False
        
    def set_permission(self, permission=None):
        """Cause the command to require a specific permission from the command 
        user.
        
        Arguments:
        permission (Keyword Default: None) - the name of the permission 
        required.
        
        """
        self.requires_auth = True
        self.permission = permission
        
class ClientSayCommand(ClientConsoleCommand):
    """Manage a client say command through a custom say filter in order to 
    correctly handle args encapsulated by "".
    
    """
    
    BASENAME = "saycommand"
    INFORMER = staticmethod(tools.tell)
    MAPPING = _command_manager.client_say_commands
    VERBOSE_NAME = "Client Say Command"
    
class Parameter(object):
    """Create a parameter that can be used by several commmands."""

    def __init__(self, basename, description=None):
        """Instantiate a Parameter object.
        
        Arguments:
        basename - the basename to identify the parameter by.
        description - a description of the values accepted.
        
        """
        self.basename = basename
        self.description = description
        self.transformer = None
        
    def set_transformer(self, transformer):
        self.transformer = transformer
        
def _server_command():
    command_word = es.getargv(0)
    command = _command_manager.server_commands[command_word]
    if not command.enabled:
        command._disabled(command_word)
        return
    args = [es.getargv(arg) for arg in xrange(1, es.getargc())]
    try:
        args = _get_transformed_args(command._callback, args)
    except ArgumentCountError:
        command._invalid_syntax()
    except ArgumentValueError, (parameter, value):
        command._invalid_argument(parameter, value)
    else:
        command._callback.callable_name(command.INFORMER, *args)
        
es.addons.registerBlock("_commands", "_server_command", _server_command)

class ServerCommand(_Command):
    """Manage a client console command, wrapping es.regcmd"""
    
    BASENAME = "command"
    INFORMER = staticmethod(tools.output)
    INTERNAL_CALLBACK = _server_command
    MAPPING = _command_manager.server_commands
    REGISTER_CALLABLE = staticmethod(es.regcmd)
    VERBOSE_NAME = "Server Command"
    
    def __init__(self, *args, **kwargs):
        """Instantiate a ServerCommand object.
        
        Arguments:
        args - the positional arguments defined by the parent class.
        kwargs - the keyword arguments defined by the parent class.
        
        """
        super(ServerCommand, self).__init__(*args, **kwargs)
        self.enabled = False
        
    def _disabled(self, name):
        self.INFORMER("Unknown command \"%s\"" % name)
        
    def _check(self, name):
        if name in self.MAPPING and self.enabled:
            raise CommandsError("%s is already registered as a %s" 
                                %(name, self.VERBOSE_NAME))
        return True
         
    def register(self, *args, **kwargs):
        """Register a command name to trigger the command's callback.
        
        Arguments:
        args - the positional arguments defined by the parent class.
        kwargs - the keyword arguments defined by the parent class.
        
        """
        super(ServerCommand, self).register(*args, **kwargs)
        self.enabled = True
        
    def unregister(self, *args, **kwargs):
        """Unregister all registered command names.
        
        Arguments:
        args - the positional arguments defined by the parent class.
        kwargs - the keyword arguments defined by the parent class.
        
        """
        super(ServerCommand, self).unregister(*args, **kwargs)
        self.enabled = False
        
def _get_transformed_args(callback, args):
    total_args = len(args)
    total_parameters = len(callback.parameters)
    if (total_args > total_parameters or total_args < total_parameters and 
        callback.parameters[total_args] not in callback.defaults):
        raise ArgumentCountError()
    transformed_args = []
    for index, arg in enumerate(args):
        parameter = callback.parameters[index]
        if parameter.transformer is not None:
            try:
                arg = parameter.transformer(arg)
            except ArgumentValueError:
                raise ArgumentValueError(parameter, arg)
        transformed_args.append(arg)
    remaining_args = total_parameters - total_args
    if remaining_args:
        transformed_args.extend(callback.defaults[parameter] 
                                for parameter in 
                                callback.parameters[-remaining_args: ])
    return transformed_args
    
def _say_filter(user_id, text, teamonly):
    stripped_text = text.strip("\"")
    command_name = stripped_text.partition(" ")[0]
    if not command_name in _command_manager.client_say_commands:
        return (user_id, text, teamonly)
    command = _command_manager.client_say_commands[command_name]
    if (command.requires_auth and 
        not admins.is_user_authorised(es.getplayersteamid(user_id), 
                                      command.permission)):
        command._no_auth(user_id)
        return (False, False, False)
    try:
        args = shlex.split(stripped_text)
    except ValueError:
        # If the string does not have a trailing ", it needs appending.
        stripped_text += "\""
        args = shlex.split(stripped_text)
    del args[0]
    try:
        args = _get_transformed_args(command._callback, args)
    except ArgumentCountError:
        command._invalid_syntax(user_id)
    except ArgumentValueError, (parameter, value):
        command._invalid_argument(parameter, value, user_id=user_id)
    else:
        command._callback.callable_name(command.INFORMER, user_id=user_id, 
                                       *args)
    return (False, False, False)
    
es.addons.registerSayFilter(_say_filter)