# <path to game directory>/addons/eventscripts/_engines/python/Lib/esutils/
# players.py
# by Adam Cunnington

"""Provide a simple API to access player attributes in a single uniform 
manner. Also provide a highly versatile filter system to get a set of player 
objects.

"""

import itertools
import functools
import math

from esutils import tools, weapons
import es

__all__ = (
    "UNASSIGNED", 
    "SPECTATOR", 
    "TERRORIST", 
    "COUNTER_TERRORIST", 
    "PlayersError",  
    "Player", 
    "all_players", 
)

# To avoid passing meaningless integers throughout the library and usage, 
# declare them once as constants.
UNASSIGNED = 0
SPECTATOR = 1
TERRORIST = 2
COUNTER_TERRORIST = 3

class PlayersError(Exception):
    """General error encountered relating to admin.libraries.players"""
    
class __PlayerAttribute(object):
    """A common player information type that is obtained through 
    es.createplayerlist(<userid>)[<userid>]. These items are read-only.
    
    """

    def __init__(self, name):
        """Instantiate a _PlayerAttribute object.
        
        Arguments:
        name - the key value for the information item.
        
        """
        self.name = name
        
    def get(self, player):
        """Return the value of the player's attribute. If it has been obtained 
        before and the attribute values don't require updating, return a 
        cached value for speed. This method should generally only be 
        implicitly called by python.
        
        Arguments:
        player - the relevant player object.
        
        """
        if player.attributes is None or player.auto_update_attributes:
            player.update_attributes()
        return player.attributes[self.name]
        
class __PlayerEntity(object):
    """A common player information type that is obtained 
    es.getplayerprop(<userid>, <entity property>). These items can be writable 
    as well as readable.
    
    """

    def __init__(self, name):
        """Instantiate a _PlayerAttribute object.
        
        Arguments:
        name - the full entity property name for the information item.
        
        """
        self.name = name
        
    def get(self, user_id):
        """Return the value of the player's entity property. Cacheing is not 
        possible and it's also unnecessary as es.getplayerprop is fast. This 
        method should generally only be implicitly called by python.
        
        Arguments:
        user_id - the relevant user ID.
        
        """
        return es.getplayeprop(user_id, self.name)
        
    def set(self, user_id, value):
        """Set the value of the player's entity property. This 
        method should generally only be implicitly called by python.
        
        Arguments:
        user_id - the relevant user ID.
        value - the new value to set the property at.
        
        """
        es.setplayerprop(user_id, self.name, value)
        
def _address_transformer(value):
    hostname, port = value.split(":")
    return (hostname, int(port))
    
def _bit_transformer(value):
    return bool(value & 1)
    
def _calculate_colour(user_id, value):
    r, g, b, a = value
    colour = r + g << 8 + b << 16 + a << 24
    if colour >= 2 ** 31: 
        colour -= 2 ** 32
    return colour
    
def _frozen_transformer(value):
    return value not in (0, 4, 8)
    
def _get_colour(user_id):
    colour = es.getplayerprop(user_id, "CBaseEntity.m_clrRender")
    return (colour & 0xff, (colour & 0xff00) >> 8, 
            (colour & 0xff0000) >> 16, (colour & 0xff000000) >> 24)
            
def _get_eye_location(user_id):
    return tuple(es.getplayerprop(self.userid, 
                                  "CBasePlayer.localdata.m_vecViewOffset[%s]" 
                                  % index) + coordinate 
                 for index, coordinate 
                 in enumerate(es.getplayerlocation(user_id)))
                 
def _get_flame_indexes(user_id):
    for index in es.createentityindexlist("entityflame"):
        ownership = es.getindexprop(index, "CEntityFlame.m_hEntAttached")
        if onwership == self.handle:
            yield index
            
def _get_language(user_id):
    return es.getclientvar(user_id, "language")
    
def _get_melee(player):
    return _get_player_weapon(player.handle, 3)
    
def _get_player_weapon(handle, slot):
    for index in weapons.indexes_by_slot(slot):
        if es.getindexprop(index, "CBaseCombatWeapon.m_hOwner") == handle:
            return weapons.Weapon(index)
            
def _get_primary(player):
    return _get_player_weapon(player.handle, 1)
    
def _get_secondary(player):
    return _get_player_weapon(player.handle, 2)
    
def _get_view_angle(user_id):
    eye_angle0, eye_angle1 = _get_view_angles(user_id)
    if eye_angle1 < 0:
        eye_angle1 += 360
    roll = es.getplayerprop(user_id, "CBaseEntity.m_angRotation").split(",")[2]
    return (eye_angle0, eye_angle1, float(roll))
    
def _get_view_angles(user_id):
    return (es.getplayerprop(user_id, "CCSPlayer.m_angEyeAngles[%s]" % index)
            for index in range(1))
    
def _get_view_coordinates(user_id):
    model = "props_c17/tv_monitor01_screen.mdl"
    es.server.cmd("es_xprop_dynamic_create %s %s" %(user_id, model))
    last_give = int(es.ServerVar("eventscripts_lastgive"))
    location = es.getindexprop(last_give, "CBaseEntity.m_vecOrigin")
    es.server.insertcmd("es_xremove %s" % last_give)
    return location
    
def _get_view_vector(user_id):
    eye_angle0, eye_angle1 = _get_view_angles(user_id)
    eye1 = math.radians(eye_angle1)
    return (math.cos(eye1), math.sin(eye1), 
            math.sin(math.radians(eye_angle0)) * -1)
            
def _get_weapon(player):
    if player.attributes is None or player.auto_update_attributes:
        player.update_attributes()
    weapon = player.attributes["weapon"]
    for index in weapons.weapon_type_by_name(weapon).indexes:
        ownership = es.getindexprop(index, "CBaseCombatWeapon.m_hOwner")
        if ownership == player.handle:
            return weapons.Weapon(index)
            
def _get_weapon_colour(player):
    index = player.weapon.index
    colour = es.getindexprop(index, "CBaseEntity.m_clrRender")
    return tuple(int(component) for component in 
                 (colour & 0xff, (colour & 0xff00) >> 8, 
                 (colour & 0xff0000) >> 16, (colour & 0xff000000) >> 24))
                 
def _has_c4(player):
    for index in weapons.weapon_type_by_name("weapon_c4").indexes:
        ownership = es.getindexprop(index, "CBaseCombatWeapon.m_hOwner")
        if ownership == player.handle:
            return True
    return False
    
def _immortal_transformer(value):
    return value == 0
    
def _jetpack_transformer(value):
    return value == 4
    
def _is_burning(user_id):
    return bool(_get_flame_indexes(user_id))
    
def _modulus_transformer(value):
    return bool(value % 2)
    
def _no_block_transformer(value):
    return value == 2
    
def _no_clip_transformer(value):
    return value == 8
    
def _player_property(getter, get_transformer=None, pass_instance_get=False, 
                    setter=None, set_transformer=None, 
                    pass_instance_set=False):
    """Create an instance property representing specific information about a 
    player. Common information types include player attributes and player 
    entity propeties but custom items are also possible.
    
    Arguments:
    getter - the callable to be used to return the value when the property is 
    queried.
    get_transformer (Keyword Default: None) - a callable that return values 
    pass through before they reach the user. This should be predominantly used 
    to ensure the end type is valid by raising an exception if not.
    pass_instance_get (Keyword Default: False) - whether or not the player 
    instance should be passed to the getter callable instead of the user's ID.
    setter (Keyword Default: None) - the callable to be used to set the value 
    when the property is queried.
    set_transformer (Keyword Default: None) - a callable that return values 
    pass through before they reach the setter callable. This should be 
    predominantly used to ensure the end type is valid by raising an exception 
    if not.
    pass_instance_set (Keyword Default: False) - whether or not the player 
    instance should be passed to the setter callable instead of the user's ID.
    
    """
    def get(self):
        if not pass_instance_get:
            parameter = self.user_id
        else:
            parameter = self
        if get_transformer is None:
            return getter(parameter)
        return get_transformer(getter(parameter))
        
    def set(self, value):
        if not pass_instance_set:
            parameter = self.user_id
        else:
            parameter = self
        if set_transformer is None:
            setter(parameter, value)
            return
        setter(parameter, set_transformer(value))
        
    if setter is None:
        return property(get)
    return property(get, set)
    
def _set_colour(user_id, value):
    es.setplayerprop(user_id, "CBaseEntity.m_nRenderMode", 
                     es.getplayerprop(user_id, "CBaseEntity.m_nRenderMode") 
                     | 1)
    es.setplayerprop(user_id, "CBaseEntity.m_nRenderFX", 
                     es.getplayerprop(user_id, "CBaseEntity.m_nRenderFX") 
                     | 256)
    es.setplayerprop(user_id, "CBaseEntity.m_clrRender", 
                     _calculate_colour(user_id, value))
                     
def _set_model(user_id, value):
    colour = _get_colour(user_id)
    es.setplayerprop(user_id, "CBaseEntity.m_nModelIndex", 
                     es.precachemodel("models/%s" % value))
    _set_colour(user_id, colour)
    
def _set_weapon_colour(player, value):
    index = player.weapon.index
    es.setindexprop(index, "CBaseEntity.m_nRenderMode", 
                    es.getindexprop(index, "CBaseEntity.m_nRenderMode") | 1)
    es.setindexprop(index, "CBaseEntity.m_nRenderFX", 
                    es.getindexprop(index, "CBaseEntity.m_nRenderFX") | 256)
    es.setindexprop(index, "CBaseEntity.m_clrRender", 
                    calculate_colour(player.user_id, value))
                    
def _str_transformer(value):
    return bool(int(value))
    
def _set_view_coordinates(user_id, value):
    x, y, z = (old - new for old, new in zip(_get_eye_location(), value))
    new_y = math.degrees(math.atan(y / x))
    if x < 0:
        new_y += 180
    elif y < 0:
        new_y += 360
    x = 0 - math.degrees(math.atan(z / math.sqrt(pow(y, 2) + pow(x, 2))))
    es.setang(user_id, x, y, _get_view_angle[2])
    
def _set_weapon(user_id, value):
    es.sexec(user_id, "use %s" % value)
    
def _transform_immortal(value):
    if value:
        return 0
    return 512
    
def _transform_jetpack(value):
    if value:
        return 4
    return 0
    
def _transform_no_block(value):
    if value:
        return 2
    return 0
    
def _transform_no_clip(value):
    if value:
        return 8
    return 0
    
def _transform_location(value):
    return "".join(map(str, value))
    
_armor = _PlayerEntity("CCSPlayer.m_ArmorValue")
_cash = _PlayerEntity("CCSPlayer.m_iAccount")
_defuser = _PlayerEntity("CCSPlayer.m_bHasDefuser")
_ducked = _PlayerEntity("CBasePlayer.localdata.m_Local.m_bDucked")
_flash_alpha = _PlayerEntity("CCSPlayer.m_flFlashMaxAlpha")
_flash_bang = _PlayerEntity("CBasePlayer.localdata.m_iAmmo.012")
_flash_duration = _PlayerEntity("CCSPlayer.m_flFlashDuration")
_he_grenade = _PlayerEntity("CBasePlayer.localdata.m_iAmmo.011")
_health = _PlayerEntity("CBasePlayer.m_iHealth")
_helmet = _PlayerEntity("CCSPlayer.m_bHasHelmet")
_immortal = _PlayerEntity("CBasePlayer.m_lifeState")
_in_rescue_zone = _PlayerEntity("CCSPlayer.m_bInHostageRescueZone")
_move_type = _PlayerEntity("CBaseEntity.movetype")
_nightvision = _PlayerEntity("CCSPlayer.m_bHasNightVision")
_nightvision_on = _PlayerEntity("CCSPlayer.m_bNightVisionOn")
_no_block = _PlayerEntity("CBaseEntity.m_CollisionGroup")
_smoke_grenade = _PlayerEntity("CBasePlayer.localdata.m_iAmmo.013")
_speed = _PlayerEntity("CBasePlayer.localdata.m_flLaggedMovementValue")

class Player(object):
    """Effectively create data store for information about a specific player. 
    Each attribute can be queried and a dynamic value will be returned. Some 
    attributes are also writable.
    
    """

    address = _player_property(_PlayerAttribute("address").get, 
                               _address_transformer, True)
    armor = _player_property(_armor.get, setter=_armor.set, 
                             set_transformer=int)
    bot = _player_property(es.isbot, bool)
    burning = _player_property(_is_burning)
    c4 = _player_property(_has_c4, pass_instance_get=True)
    can_plant = _player_property(_PlayerEntity("CCSPlayer.m_bInBombZone").get, 
                                 _modulus_transformer)
    cash = _player_property(_cash.get, setter=_cash.set, set_transformer=int)
    colour = _player_property(_get_colour, setter=_set_colour, 
                              set_transformer=tuple)
    connected = _player_property(tools.is_connected, bool)
    dead = _player_property(_PlayerEntity("CBasePlayer.pl.deadflag").get, bool)
    deaths = _player_property(_PlayerAttribute("deaths").get, int, 
                              pass_instance_get=True)
    defuser = _player_property(_defuser.get, _modulus_transformer, 
                               _defuser.set, bool)
    defusing = _player_property(_PlayerEntity("CCSPlayer.m_bIsDefusing").get, 
                                _modulus_transformer)
    ducked = _player_property(_ducked.get, _modulus_transformer)
    flash_alpha = _player_property(_flash_alpha.get, setter=_flash_alpha.set, 
                                   set_transformer=int)
    flash_bang = _player_property(_flash_bang.get, setter=_flash_bang.set, 
                                  set_transformer=int)
    flash_duration = _player_property(_flash_duration.get, 
                                      setter=_flash_duration.set, 
                                      set_transformer=int)
    frozen = _player_property(_move_type.get, _frozen_transformer)
    handle = _player_property(es.getplayerhandle)
    he_grenade = _player_property(_he_grenade.get, setter=_he_grenade.set, 
                                  set_transformer=int)
    health = _player_property(_health.get, setter=_health.set, 
                              set_transformer=int)
    helmet = _player_property(_helmet.get, _modulus_transformer, _helmet.set, 
                              bool)
    immortal = _player_property(_immortal.get, _immortal_transformer, 
                                setter=_immortal.set, 
                                set_transformer=_transform_immortal)
    in_buy_zone = _player_property(_PlayerEntity("CCSPlayer.m_bInBuyZone").get, 
                                   _modulus_transformer)
    in_rescue_zone = _player_property(_in_rescue_zone.get, 
                                      _modulus_transformer)
    in_vehicle = _player_property(_PlayerAttribute("isinavehicle").get, 
                                  _str_transformer, True)
    index = _player_property(_PlayerAttribute("index").get, int, True)
    jetpack = _player_property(_move_type.get, _jetpack_transformer, 
                               setter=_move_type.set, 
                               set_transformer=_transform_jetpack)
    kills = _player_property(_PlayerAttribute("kills").get, int, True)
    language = _player_property(_get_language)
    location = _player_property(es.getplayerlocation, pass_instance_get=True, 
                                setter=es.setpos, 
                                set_transformer=_transform_location)
    melee = _player_property(_get_melee, pass_instance_get=True)
    model = _player_property(_PlayerAttribute("model").get, 
                             pass_instance_get=True, setter=_set_model, 
                             set_transformer=str)
    name = _player_property(_PlayerAttribute("name").get, 
                            pass_instance_get=True)
    nigthvision = _player_property(_nightvision.get, _modulus_transformer, 
                                   _nightvision.set, bool)
    nigthvision_on = _player_property(_nightvision_on.get, 
                                      _modulus_transformer, 
                                      _nightvision_on.set, bool)
    no_block = _player_property(_no_block.get, _no_block_transformer, 
                                setter=_no_block.set, 
                                set_transformer=_transform_no_block)
    no_clip = _player_property(_move_type.get, _no_clip_transformer, 
                               setter=_move_type.set, 
                               set_transformer=_transform_no_clip)
    observer = _player_property(_PlayerAttribute("isobserver").get, 
                                _str_transformer, True)
    on_ground = _player_property(_PlayerEntity("CBasePlayer.m_fFlags").get, 
                                 _bit_transformer)
    packet_loss = _player_property(_PlayerAttribute("packetloss").get, int, 
                                   True)
    ping = _player_property(_PlayerAttribute("ping").get, int, True)
    primary = _player_property(_get_primary, pass_instance_get=True)
    secondary = _player_property(_get_secondary, pass_instance_get=True)
    serial_number = _player_property(_PlayerAttribute("serialnumber").get, 
                                     int, True)
    smoke_grenade = _player_property(_smoke_grenade.get, 
                                     setter=_smoke_grenade.set, 
                                     set_transformer=int)
    speed = _player_property(_speed.get, setter=_speed.set, 
                             set_transformer=float)
    steam_id = _player_property(es.getplayersteamid)
    team_id = _player_property(es.getplayerteam)
    time_connected = _player_property(_PlayerAttribute("time_connected").get, 
                                      float, True)
    tv = _player_property(_PlayerAttribute("ishltv").get, _str_transformer, 
                          True)
    view_angle = _player_property(_get_view_angle, setter=es.setang, 
                                  set_transformer=tuple)
    view_coordinates = _player_property(_get_view_coordinates, 
                                        setter=_set_view_coordinates, 
                                        set_transformer=tuple)
    view_vector = _player_property(_get_view_vector)
    weapon = _player_property(_get_weapon, pass_instance_get=True, 
                              setter=_set_weapon, set_transformer=str)
    weapon_colour = _player_property(_get_weapon_colour, 
                                     pass_instance_get=True, 
                                     setter=_set_weapon_colour, 
                                     set_transformer=tuple, 
                                     pass_instance_set=True)
                                     
    def __init__(self, user_id):
        """Instantiate a Player object.
        
        Arguments:
        user_id - the user ID of the player to store information about.
        
        """
        if not tools.is_connected(user_id):
            raise PlayersError("user ID %s does not exist on the server." 
                               % user_id)
        self.user_id = user_id
        self.attributes = None
        self.auto_update_attributes = False
        
    def ban_address(self, time, reason):
        """Ban and kick a player by ip.
        
        Arguments:
        time - how long the player should be banned by ip for.
        reason - why they should be banned.
        
        """
        es.server.queuecmd("banip %s %s"  %(self.attributes["address"], time))
        self.kick(reason)
        
    def ban_steam_id(self, time, reason):
        """Ban and kick a player by their steam id.
        
        Arguments:
        time - how long the player should be banned by steam id for.
        reason - why they should be banned.
        
        """
        es.server.queuecmd("banid %s %s"  %(self.user_id, time))
        self.kick(reason)
        
    def burn(self):
        """Set the player on fire and cause them to take damage."""
        es.server.cmd("es_xfire %s !self ignite" % self.user_id)
        
    def change_team(self, team):
        """Change the user's team, this will cause them to die.
        
        Arguments:
        team - the integer of the team to change the user to. Rather than 
        directly passing an integer, players.SPECTATOR, players.TERRORIST 
        and players.COUNTER_TERRORIST should be passed.
        
        """
        es.server.queuecmd("es_xchangeteam %s %s" %(self.user_id, team))
        
    def closest_player(self, opponent_only=False):
        """Return the user ID of the closest player.
        
        Arguments:
        opponent_only (Keyword Default: False) - whether or not only opponent 
        players should be taken into account when finding the closest player.
        
        """
        def get_targets(team=None):
            players = map(Player, es.getUseridList())
            if team is not None:
                players = (player for player in players 
                           if player.team_id == 5 - team)
            for player in players:
                if player.user_id == self.user_id:
                    continue
                if not player.dead:
                    yield player
                    
        if self.team_id not in (2, 3) or self.dead:
            return
        if opponent_only:
            targets = get_targets(5 - self.team_id)
        else:
            targets = get_targets()
        if not targets:
            return
        location = self.location
        return min((target.user_id, distance(location, target.location)) 
                   for target in targets)
                   
    def distance(self, x, y, z):
        """Return the player's distance from a location.
        
        Arguments:
        x - the x coordinate.
        y - the y coordinate.
        z - the z coordinate.
        
        """
        coordinates = (old - new for old, new in zip(_get_eye_location(), 
                                                     (x, y, z)))
        return math.sqrt(sum(pow(coordinate, 2) for coordinate in coordinates))
        
    def drop(self, weapon_name=None):
        """Force the user to drop their weapon. If a specific weapon is 
        passed, cause them to switch to it first.
        
        Arguments:
        weapon_name (Keyword Default: None) - the full name of the weapon to 
        drop.
        
        """
        if weapon_name is not None:
            _set_weapon(user_id, weapon_name)
        es.sexec(self.user_id, "drop")
        
    def extinguish(self):
        """Stop the player from being on fire and taking damage from 
        burning.
        
        """
        for index in _get_flame_indexes(self.user_id):
            es.setindexprop(index, "CEntityFlame.m_flLifetime", 0)
            
    def get_ammo(self, weapon_name):
        """Return the ammo amount in the player's weapon.
        
        Arguments:
        weapon_name - the full name of the weapon to find the amount of ammo 
        of.
        
        """
        return es.getplayerprop(self.user_id, 
                                weapons.weapon_type_by_name(weapon_name))
                                
    def give(self, entity_classname):
        """Give the player an entity by it's classname.
        
        Arguments:
        entity_classname - the full keyvalue name of the weapon as it appears 
        in es.createentitylist.
        
        """
        es.delayed(0, "es_xgive %s %s" %(self.user_id, entity_classname))
        
    def flash(self, alpha, duration):
        """Forcefully flash a player a certain brightness for a given time.
        
        Arguments:
        alpha - the brightness to flash them.
        duration - the time in seconds to flash them for including the 
        brightness fade time.
        
        """
        self.flash_alpha = 0
        self.flash_duration = 0
        self.flash_alpha = alpha
        self.flash_duration = duration
        
    def freeze(self):
        """Stop a player from being able to move. If they were moving before, 
        footstep sounds will continue to emit. They will still be treated as 
        if they were walking, i.e weapon fire will resemble any inaccuracies.
        
        """
        es.setplayerprop(self.user_id, "CBaseEntity.movetype", 2)
        
    def kick(self, reason):
        """Forcefully remove a player from the server.
        
        Arguments:
        reason - why the should be kicked.
        
        """
        es.server.queuecmd("kickid %s %s" %(self.user_id, reason))
        
    def kill(self):
        """Force a player to commit suicide and die."""
        es.sexec(self.user_id, "kill")
        
    def push(self, horizontal, vertical, vertical_override=False):
        """Forcefully move a player in a given direction.
        
        Keyword Arguments:
        horizontal - the value to push them in the x axis direction.
        vertical - the value to push them in the y axis direction.
        vertical_override (Keyword Default: False) - whether or not the 
        vertical value should take precedence in calculating the vector to 
        move the player through.
        
        """
        x, y, z = _get_view_vector(self.user_id)
        horizontal_x = horizontal * x
        horizontal_y = horizontal * y
        if not vertical_override:
            vertical *= z
        es.setplayerprop(self.user_id, 
                         "CBasePlayer.localdata.m_vecBaseVelocity", "%f,%f,%f" 
                                                                    %(x, y, z))
                                                                    
    def remove_weapon(self, weapon_name):
        """Forcefully remove a specific weapon from the player.
        
        Arguments:
        weapon_name - the full name of the weapon to be removed.
        
        """
        for index in weapons.weapon_type_by_name(weapon_name).indexes:
            ownership = es.getindexprop(self.user_id, 
                                        "CBaseCombatWeapon.m_hOwner")
            if ownership == self.handle:
                es.server.insertcmd("es_xremove %s" % index)
                return
        raise
        
    def set_ammo(self, weapon_name, value):
        """Set the ammo of a player's weapon.
        
        Arguments:
        weapon_name - the full name of the weapon to have it's ammo set.
        value - the new value to set the ammo at.
        
        """
        weapon = weapons.weapon_type_by_name(weapon_name)
        es.setplayerprop(self.user_id, weapon.ammo_property, value)
        
    def strip(self):
        """Strip a player of all their weapons."""
        strip_entity = "player_weaponstrip"
        self.give(strip_entity)
        es.delayed(0, "es_xfire %s %s strip" %(self.user_id, strip_entity))
        
    def unfreeze(self):
        """Unfreeze a player and allow them to move again."""
        es.setplayerprop(self.user_id, "CBaseEntity.movetype", 0)
        
    def update_attributes(self):
        """Update the cached attribute values for the player."""
        self.attributes = es.createplayerlist(self.user_id)[self.user_id]
        
# Create a single public way of obtaining a base object of all players which 
# can be further filtered.
all_players = functools.partial(tools.Filter, (Player(user_id) for user_id 
                                               in es.getUseridList()))